"""Size limit global variables."""


MAX_ROWS = 2000  # Maximum number of sents/rows
MAX_SENT_LENGTH = 100  # Maximum words in each sentence
MAX_WORD_LENGTH = 100  # Maximum length of each word

MAX_CATEGORIES = 20  # Maximum number of categories in a categorical field.



""" A mention, which is stored by the :class:`~puggle.Annotation.Annotation`
class."""


class Mention:
    """A single entity mention. Captures the start, end, tokens and
    labels of the mention."""

    def __init__(
        self,
        start: int,
        end: int,
        tokens: list,
        label: str,
        mention_id: int,
    ):
        """Create a new Mention.

        Args:
            start (int): The index of the first token of the mention.
            end (int): The index of the last token of the mention.
            tokens (list): The list of tokens appearing in the mention.
            label (str): The label of the mention.
            mention_id (int): The index of this mention with respect to the
               Document in which it appears.
        """
        super().__init__()
        if start == end:
            raise ValueError(
                "Mention start index cannot be the same as its end index."
            )
        self.start = start
        self.end = end
        self.tokens = tokens
        self.label = label
        self.mention_id = mention_id

    def to_dict(self):
        """Return a dictionary representation of this mention.
        Don't include the mention_id as it is not useful here - it is only
        used when creating Relation objects between Mentions.

        Returns:
            dict: The mention as a dictionary.
        """
        return {k: v for k, v in self.__dict__.items() if k != "mention_id"}

    def __repr__(self):
        return f"({' '.join(self.tokens)} [{self.label}]) (start: {self.start}, end: {self.end})"



"""Tools for manipulating Puggle datasets. When this library is imported,
the functions will be added to the Dataset class."""
from puggle import Dataset
from puggle.logger import logger


def drop_entity_class(self: Dataset, entity_class: str):
    """
    :bdg-success-line:`Manipulation`
    Remove any instances of the given entity class from the mentions of this
    Dataset.
    Also remove all Relations referencing the deleted mentions.

    Args:
        entity_class (str): The entity class to remove.
    """
    dataset = self
    n_removed_e = 0
    n_removed_r = 0
    for doc in dataset.documents:
        a = doc.annotation

        updated_mentions = list(
            filter(lambda m: m.label != entity_class, a.mentions)
        )
        n_removed_e += len(a.mentions) - len(updated_mentions)
        a.mentions = updated_mentions

        # Remove connected relations
        updated_relations = list(
            filter(
                lambda r: ((r.start in a.mentions) or (r.end in a.mentions)),
                a.relations,
            )
        )
        n_removed_r += len(a.relations) - len(updated_relations)
        a.relations = updated_relations

    logger.info(
        f'Drop Entity class "{entity_class}": removed {n_removed_e} entities '
        f"and {n_removed_r} connected relations."
    )


def drop_relation_class(self: Dataset, relation_class: str):
    """
    :bdg-success-line:`Manipulation`
    Remove any instances of the given relation class from the relations of
    this Dataset.

    Args:
        relation_class (str): The relation class to remove.
    """
    dataset = self
    n_removed = 0
    for doc in dataset.documents:
        a = doc.annotation
        n_removed += sum(r.label == relation_class for r in a.relations)
        a.relations = list(
            filter(lambda r: r.label != relation_class, a.relations)
        )

    logger.info(
        f'Drop Relation class "{relation_class}": removed {n_removed} relations.'
    )


def convert_entity_class(self: Dataset, original_ec: str, modified_ec: str):
    """
    :bdg-success-line:`Manipulation`
    Convert the given entity class from one label to another across the
    entire Dataset.

    Args:
        original_ec (str): The entity class to change.
        modified_ec (str): The entity class to change to.
    """
    n_modified = 0
    for doc in self.documents:
        a = doc.annotation
        for m in a.mentions:
            if m.label == original_ec:
                m.label = modified_ec
                n_modified += 1
    logger.info(
        f'Convert Entity class "{original_ec}" -> "{modified_ec}": '
        f" modified {n_modified} entities."
    )


def convert_relation_class(self: Dataset, original_rc: str, modified_rc: str):
    """
    :bdg-success-line:`Manipulation`
    Convert the given relation class from one label to another across the
    entire Dataset.

    Args:
        original_ec (str): The relation class to change.
        modified_ec (str): The relation class to change to.
    """
    n_modified = 0
    for doc in self.documents:
        a = doc.annotation
        for r in a.relations:
            if r.label == original_rc:
                r.label = modified_rc
                n_modified += 1
    logger.info(
        f'Convert Relation class "{original_rc}" -> "{modified_rc}": '
        f"modified {n_modified} relations."
    )


def flatten_all_entities(self: Dataset):
    """
    :bdg-success-line:`Manipulation`
    Flatten all entities, i.e. resolve all hierarchical entities to their
    base class. For example,
    ["state/desirable"] becomes ["state"], etc.
    """
    n_modified = 0
    for doc in self.documents:
        a = doc.annotation
        for m in a.mentions:
            m.label = m.label.split("/")[0]
        n_modified += len(a.mentions)
    logger.info(
        f"Successfully flattened all {n_modified} entities in dataset."
    )


def flatten_all_relations(self: Dataset):
    """
    :bdg-success-line:`Manipulation`
    Flatten all relations, i.e. resolve all hierarchical relations to their
    base class. For example,
    ["state/desirable"] becomes ["state"], etc.
    """
    n_modified = 0
    for doc in self.documents:
        a = doc.annotation
        for r in a.relations:
            r.label = r.label.split("/")[0]
        n_modified += len(a.relations)
    logger.info(
        f"Successfully flattened all {n_modified} relations in dataset."
    )


Dataset.drop_entity_class = drop_entity_class
Dataset.drop_relation_class = drop_relation_class
Dataset.convert_entity_class = convert_entity_class
Dataset.convert_relation_class = convert_relation_class
Dataset.flatten_all_entities = flatten_all_entities
Dataset.flatten_all_relations = flatten_all_relations



"""Functions for working with Datasets in Puggle."""
from .sampling import (
    random_sample,
    smart_sample,
)
from .statistics import get_unique_tokens_count, get_entity_label_counts
from .manipulation import drop_entity_class, drop_relation_class



"""Functions for generating samples from Puggle Datasets."""
import random
from puggle import Dataset, Document
from puggle.logger import logger


def random_sample(self: Dataset, num_records: int) -> Dataset:
    """
    :bdg-primary-line:`sampling`
    Run a 'random sample' over the given dataset to return a new
    Dataset with num_records documents.

    Args:
        self (Dataset): The dataset to sample.
        num_records (int): The number of documents that should appear in the
           output.
    """
    output_dataset = Dataset()
    sample = random.sample(self.documents, num_records)
    for doc in sample:
        output_dataset.add_document(doc)
    return output_dataset


def random_split(self: Dataset) -> (Dataset, Dataset, Dataset):
    """
    :bdg-primary-line:`sampling`
    Randomly split this dataset into 3 datasets - 80% train, 10% dev, 10% test.

    Returns:
        Dataset, Dataset, Dataset: The train, dev and test datasets.
    """

    rs = random.sample(self.documents, len(self.documents))
    train = Dataset()
    dev = Dataset()
    test = Dataset()
    for doc in rs[: int(len(rs) * 0.8)]:
        train.add_document(doc)
    for doc in rs[int(len(rs) * 0.8) : int(len(rs) * 0.9)]:
        dev.add_document(doc)
    for doc in rs[int(len(rs) * 0.9) :]:
        test.add_document(doc)

    logger.info(
        "Randomly split dataset into %s train, %s dev and %s test."
        % (len(train.documents), len(dev.documents), len(test.documents))
    )

    return train, dev, test


def smart_sample(self: Dataset, num_records: int, num_samples: int) -> Dataset:
    """
    :bdg-primary-line:`sampling`

    .. warning::

        This function is experimental - it works, but hasn't been tested.

    Run a 'smart sample' on the given dataset to return a new Dataset with
    num_records documents.
    Repeat the sampling process num_samples times and select the best
    example.
    The algorithm aims to maximise the number of different tokens, entity
    classes and relation classes in the sample.

    Args:
        num_records (int): The number of documents that should appear in the
           output.
        num_samples (int): Number of times to repeat the process. The idea is
           that the more samples are run, the more likely it is that the
           function will generate a better quality sample.
    """
    dataset = self
    documents = dataset.documents

    # Repeat num_samples times to create a list of num_samples different
    # 'sample sets'. The best of these will be chosen as the final output.
    sample_sets = []
    for sample_index in range(num_samples):
        logger.info(f"Running sample {sample_index}...")
        # Begin by selecting a random document.
        start_doc = random.choice(documents)
        sample_set = [start_doc]

        # For num_records - 1 times, calculate the scores of each document
        # against the current sample set in this iteration.
        # This will initially be just one document (the randomly selected
        # one from earlier), and will be iteratively built up 1 document
        # at a time.
        for _ in range(num_records - 1):
            scored_documents = []
            for doc in documents:
                if doc in sample_set:
                    continue

                # Use the scoring function to determine the score of this
                # document. Higher scores are better.
                score = _calculate_document_score(doc, sample_set)
                scored_documents.append((doc, score))

            scored_documents.sort(key=lambda x: x[1], reverse=True)
            sample_set.append(scored_documents[0][0])

        sample_sets.append(sample_set)

    output_dataset = Dataset()

    # Now that we have num_samples documents in the sample set,
    # calculate the average score of every document in that set against
    # every other document in the set.
    # The average score determines how 'good' our sample set is.
    scored_sets = []
    for sample_set in sample_sets:
        average_score = _calculate_sample_quality(
            sample_set, dataset.documents
        )
        scored_sets.append((sample_set, average_score))

    # Sort these sample sets, taking the one with the lowest score as the
    # final output dataset.
    scored_sets.sort(key=lambda x: x[1], reverse=True)
    logger.debug(f"Average scores of each of the {num_samples} sample sets:")
    for _, score in scored_sets:
        logger.debug(score)

    logger.debug(
        "Creating final output dataset from the lowest-scored sample..."
    )
    for d in scored_sets[0][0]:
        output_dataset.add_document(d)

    return output_dataset


def _calculate_sample_quality(
    sample_dataset: list[Document], full_dataset: list[Document]
):
    """Calculate the 'similarity' of the given list of Documents using the
    3-part scoring method.

    Args:
        sample_dataset (Dataset): The sample dataset..
        full_dataset (Dataset): The full list of documents.

    Returns:
        float: The final score.
    """
    average_score = sum(
        _calculate_document_score(doc, full_dataset) for doc in sample_dataset
    ) / len(sample_dataset)
    return average_score


def _calculate_document_score(document: Document, sample_set: list[Document]):
    """Calculate the 'score' of the given document by comparing it
    to every document in the sample set so far.

    Args:
        document (Document): The document to calculate the score of.
        sample_set (Dataset): The current sample set.

    Returns:
        float: The total score.
    """

    # Token score
    # number of documents in which that token appears / number of documents
    token_score = _get_token_score(document, sample_set)

    # Entity score
    # number of documents in which that entity appears / number of documents
    entity_score = _get_entity_score(document, sample_set)

    # Relation score
    # number of documents in which that relation appears / number of documents
    relation_score = _get_relation_score(document, sample_set)

    # print(token_score, entity_score, relation_score)

    return (token_score + entity_score + relation_score) / 3


def _get_token_score(document: Document, sample_set: list[Document]):
    """Get the token score.

    Args:
        document (Document): The document to get the score for.
        sample_set (list[Document]): The other docs in the set.

    Returns:
        float: The token score.
    """

    # If there are no tokens, this doc is not useful to sample,
    # thus the score is set to the lowest possible value of 0.
    if len(document.annotation.tokens) == 0:
        return 0.0

    token_scores = []
    for t in document.annotation.tokens:
        freq = 0
        for other_doc in sample_set:
            if t in other_doc.annotation.tokens:
                freq += 1
        ts = freq / len(sample_set)
        token_scores.append(ts)
    token_score = (
        sum(token_scores) / len(token_scores) if len(token_scores) > 0 else 0
    )
    return 1 - token_score


def _get_entity_score(document: Document, sample_set: list[Document]):
    """Get the entity score.

    Args:
        document (Document): The document to get the score for.
        sample_set (list[Document]): The other docs in the set.

    Returns:
        float: The entity score.
    """

    # If there are no mentions, this doc is not useful to sample,
    # thus the score is set to the lowest possible value of 0.
    if len(document.annotation.mentions) == 0:
        return 0.0

    entity_scores = []
    for m in document.annotation.mentions:
        label = m.label

        freq = 0
        for other_doc in sample_set:
            for m in other_doc.annotation.mentions:
                if label == m.label:
                    freq += 1
                    break
        es = freq / len(sample_set)
        entity_scores.append(es)
    entity_score = (
        sum(entity_scores) / len(entity_scores)
        if len(entity_scores) > 0
        else 0
    )
    return 1 - entity_score


def _get_relation_score(document: Document, sample_set: list[Document]):
    """Get the relation score.

    Args:
        document (Document): The document to get the score for.
        sample_set (list[Document]): The other docs in the set.

    Returns:
        float: The relation score.
    """

    # If there are no relations, this doc is not useful to sample,
    # thus the score is set to the lowest possible value of 1.
    if len(document.annotation.relations) == 0:
        return 0.0

    relation_scores = []
    for r in document.annotation.relations:
        label = r.label
        freq = 0
        for other_doc in sample_set:
            for _ in other_doc.annotation.relations:
                if label == r.label:
                    freq += 1
                    break
        rs = freq / len(sample_set)
        relation_scores.append(rs)
    relation_score = (
        sum(relation_scores) / len(relation_scores)
        if len(relation_scores) > 0
        else 0
    )
    return 1 - relation_score


Dataset.random_sample = random_sample
Dataset.random_split = random_split
Dataset.smart_sample = smart_sample



"""Utility functions such as normalising annotation formats, etc."""
from typing import Dict


def validate_anns_format(anns_format: str):
    """Helper function to ensure anns_format is valid.

    Args:
        anns_format (str): The format of the annotations.

    Raises:
        ValueError: If format not valid.
    """
    valid_formats = ["quickgraph", "spert"]
    if anns_format not in ["quickgraph", "spert"]:
        raise ValueError(f"anns_format must be in {valid_formats}.")


def normalise_annotation_format(doc: Dict, anns_format: str):
    """Normalise the given document (as a dict) into a uniform format.

    Args:
        doc (Dict): The document to normalise.
        anns_format (str): The annotation format.

    Returns:
        Dict: The normalised document.
    """
    validate_anns_format(anns_format)

    if anns_format == "quickgraph":
        return _normalise_quickgraph(doc)
    return _normalise_spert(doc)


def _normalise_spert(doc: Dict):
    """Normalise the spert-formatted JSON to be parsable by puggle.

    Args:
        doc (Dict): The annotations to parse.

    Returns:
        Dict: The parsed document.
    """
    if "entities" not in doc:
        doc["entities"] = []
    if "relations" not in doc:
        doc["relations"] = []

    for m in doc["entities"]:
        m["label"] = m["type"]
        del m["type"]

    for r in doc["relations"]:
        r["start"] = r["head"]
        r["end"] = r["tail"]
        del r["head"]
        del r["tail"]

    return doc


def _normalise_quickgraph(doc: Dict):
    """Normalise the quickgraph formatted JSON to be parsable by puggle.

    Args:
        doc (Dict): The annotations to parse.

    Returns:
        Dict: The parsed document.
    """
    if "entities" not in doc:
        doc["entities"] = []
    if "relations" not in doc:
        doc["relations"] = []

    entity_idxs = {}
    for i, m in enumerate(doc["entities"]):
        m["label"] = m["label"]
        entity_idxs[m["id"]] = i
        m["end"] = m["end"] + 1
        del m["id"]

    for r in doc["relations"]:
        r["start"] = entity_idxs[r["source_id"]]
        r["end"] = entity_idxs[r["target_id"]]
        r["type"] = r["label"]
        del r["label"]
        del r["source_id"]
        del r["target_id"]
        if "id" in r:
            del r["id"]

    return doc



""" A relation, which captures the relationship between one
:class:`~puggle.Annotation.Mention` and another. It is stored by the
:class:`~puggle.Annotation.Annotation` class."""
from .Mention import Mention


class Relation:
    """A single relation. Captures the start mention, end mention, tokens and
    labels of the relation."""

    def __init__(self, start: Mention, end: Mention, label: str):
        """Create a new Relation.

        Args:
            start (Mention): The start (head) Mention.
            end (Mention): The end (tail) Mention.
            label (str): The label (type) of the relation.

        Raises:
            ValueError: If the start and end Mention is the same.
        """
        super().__init__()

        if start == end:
            raise ValueError(
                "Cannot create relation between a mention and itself"
            )

        self.start = start
        self.end = end
        self.label = label

    def to_dict(self):
        """Return a dictionary representation of this Relation.
        Convert 'start' and 'end' to the mention index of this
        relation's start and end mention.

        Returns:
            dict: Dictionary representation of this Relation.
        """
        return {
            "start": self.start.mention_id,
            "end": self.end.mention_id,
            "type": self.label,
        }

    def __repr__(self):
        return (
            f"({' '.join(self.start.tokens)})-"
            f"[{self.label}]->({' '.join(self.end.tokens)})"
        )



"""Statistics-based functions for the Dataset class."""
from puggle import Dataset


def get_unique_tokens_count(self: Dataset):
    """
    :bdg-danger-line:`Statistics`
    Return the number of unique tokens in this Dataset.

    Args:
        dataset (Dataset): The dataset to use.

    Returns:
        int: The number of unique tokens in the dataset.
    """

    seen_tokens = set()
    for d in self.documents:
        for token in d.annotation.tokens:
            seen_tokens.add(token)

    return len(list(seen_tokens))


def get_entity_label_counts(self: Dataset, document_level=False):
    """
    :bdg-danger-line:`Statistics`
    Return a sorted list of (entity_label, freq) pairs in this Dataset.
    The frequency is the number of times that entity_label has been used.

    Args:
        document_level (bool, optional): If True, the counts will be the
           number of documents in which the entity label appears, rather than
           the total frequency of that entity label.

    Returns:
        list[tuple]: A sorted list of (entity_label, freq) pairs.
    """
    counts_dict = {}
    for d in self.documents:
        seen_this_doc = set()
        for m in d.annotation.mentions:
            label = m.label
            if label not in counts_dict:
                counts_dict[label] = 0
            if label not in seen_this_doc or not document_level:
                counts_dict[label] += 1
            seen_this_doc.add(label)

    sorted_counts = sorted(
        counts_dict.items(), key=lambda x: x[1], reverse=True
    )

    return sorted_counts


Dataset.get_unique_tokens_count = get_unique_tokens_count
Dataset.get_entity_label_counts = get_entity_label_counts



"""A Dataset that stores a list of Documents.
"""
import os
import json
import csv
import logging as logger
from json import JSONDecodeError
from typing import Dict, List
from py2neo import Graph
from dotenv import load_dotenv

from .Document import Document
from .Annotation import Annotation
from .utils import (
    validate_anns_format,
    normalise_annotation_format,
)
from .logger import logger

load_dotenv()


class Dataset(object):
    """A class representing a Dataset, which stores a list of Documents.

    :var documents: A List of :class:`puggle.Document.Document` objects.
    """

    def __init__(self):
        """Create an empty Dataset. Documents may be loaded via the
        :func:`puggle.Dataset.Dataset.load_documents` function.
        """
        super().__init__()
        self.documents = []

    def load_documents(
        self,
        sd_filename: os.path = None,
        anns_filename: os.path = None,
        anns_format: str = None,
    ):
        """Load a set of documents given the filepath of the structured data
        (a .csv file), and the filepath of the annotations (a .json file).
        Documents can still be created if either one of these is not present,
        but not if both are not present.
        Each row of each file must correspond to the other, e.g. row 3 of the
        structured data csv must correspond to row 3 of the annotations json.

        Args:
            sd_filename (os.path, optional): The filepath of the structured
               data.
            anns_filename (os.path, optional): The filepath of the
               annotations.
            anns_format (str): The format of the annotations file. Can be
               either "quickgraph" or "spert".
        """
        if sd_filename is None and anns_filename is None:
            raise ValueError(
                "Either sd_filename or anns_filename (or both) must be "
                "present in order to load Documents."
            )
        if anns_filename is not None:
            validate_anns_format(anns_format)

        structured_fields = []
        annotations = []

        if sd_filename is not None:
            structured_fields = self._load_structured_data(sd_filename)
        if anns_filename is not None:
            annotations = self._load_annotations(anns_filename, anns_format)

        if all((structured_fields, annotations)) and len(
            structured_fields
        ) != len(annotations):
            raise ValueError(
                "Mismatch between the length of the structured "
                "fields dataset and the annotations dataset."
            )

        if len(structured_fields) == 0:
            structured_fields = [None] * len(annotations)

        documents = []
        for sf, ann in zip(structured_fields, annotations):
            d = Document(sf, ann)
            documents.append(d)

        self.documents += documents
        logger.info(
            f"Successfully loaded {len(documents)} documents. "
            f"Dataset now contains {len(self.documents)} documents in total."
        )

    def save_to_file(self, filename: str, output_format: str = "json"):
        """Save the documents of this dataset to the given filename.

        There are two `output_format` options to choose from: `json`
        and `quickgraph`. See the "Basic functionality" section of the
        documentation for more info.

        Args:
            filename (str): The filename to save to.
            output_format (str): The format to save to. 'json' will save as a
               json file without any special formatting. 'spert' will save it
               ready for using in SPERT. 'quickgraph' will save
               as a json file that can be loaded directly into quickgraph.
        """
        if output_format not in ["json", "spert", "quickgraph"]:
            raise ValueError(
                "Output format must be either 'json', 'spert' or 'quickgraph'"
            )

        if output_format == "json":
            with open(filename, "w", encoding="utf-8") as f:
                json.dump(
                    [doc.to_dict() for doc in self.documents], f, indent=2
                )
        elif output_format == "spert":
            spert_docs = _to_spert(self)
            with open(filename, "w", encoding="utf-8") as f:
                json.dump(spert_docs, f, indent=2)
        elif output_format == "quickgraph":
            quickgraph_docs = _to_quickgraph(self)
            with open(filename, "w", encoding="utf-8") as f:
                json.dump(quickgraph_docs, f, indent=2)

        logger.info(f"Saved dataset to %s." % filename)

    def load_into_neo4j(self, recreate=False):
        """Load the Dataset into a Neo4j database.
        Automatically creates Nodes from the entities (mentions) appearing
        in each document, and relationships between them via the Relations.

        Args:
            recreate (bool, optional): If true, the Neo4j db will be cleared
                prior to inserting the documents into it.

        Raises:
            RuntimeError: If the Neo4j server is not running.
        """

        port = os.getenv("NEO4J_PORT") if "NEO4J_PORT" in os.environ else 7687

        # Init graph
        graph = Graph(
            password=os.getenv("NEO4J_PASSWORD"),
            port=port,
        )

        # Attempt to run a query to make sure it is working
        try:
            graph.run("MATCH () RETURN 1 LIMIT 1")
        except Exception as e:
            raise RuntimeError(
                "The Neo4j graph does not appear to be running. "
                f"Please run Neo4j on port {port} to proceed."
            )

        if recreate:
            graph.run("MATCH (n) DETACH DELETE (n)")

        for i, d in enumerate(self.documents):
            # Create node for document
            cypher = f"MERGE (d:Document {{doc_idx: {i}}})"
            graph.run(cypher)

            ann = d.annotation
            # If this doc has no annotation, skip the creation of
            # entities and relationships
            if ann is None:
                continue

            # Update the Document node to copy over each of the fields
            # from the CSV dataset
            if d.fields:
                for k, v in d.fields.items():
                    val = ('"' + v + '"') if type(v) is str else v
                    cypher = (
                        f"MATCH (d:Document {{doc_idx: {i}}})\n"
                        f"SET d.{k} = {val}"
                    )
                    graph.run(cypher)

            # Do not continue building the graph if the annotation
            # has no relationships between entities
            if ann.relations is None:
                continue
            for rel in ann.relations:
                # Avoid cyclical relationships
                if rel.start.tokens == rel.end.tokens:
                    continue

                # Create relationship between head and tail
                cypher = (
                    f"MATCH (d:Document {{doc_idx: {i}}})\n"
                    f"MERGE (e1:Entity:{rel.start.label} {{name: "
                    f"\"{' '.join(rel.start.tokens)}\"}})\n"
                    f"MERGE (e2:Entity:{rel.end.label}  {{name: "
                    f"\"{' '.join(rel.end.tokens)}\"}})\n"
                    f"MERGE (e1)-[r:{rel.label}]->(e2)\n"
                    f"MERGE (e1)-[:APPEARS_IN]->(d)\n"
                    f"MERGE (e2)-[:APPEARS_IN]->(d)"
                )
                graph.run(cypher)
            if i > 0 and i % 10 == 0:
                logger.info(f"Processed {i} documents")
        logger.info("Graph creation complete.")

    def add_document(self, document: Document):
        """Add the given Document to this Dataset.

        Args:
            document (Document): The Document to add.

        """
        self.documents.append(document)

    # def create_neo4j_csvs():
    #     """A function to generate a set of CSVs to load into Neo4j via
    #     IMPORT statements (an alternative for those who want to be able
    #     to save their graph to disk somehow and import it later/elsewhere).

    #     Raises:
    #         NotImplementedError: Description
    #     """
    #     raise NotImplementedError

    def _load_structured_data(self, filename: os.path):
        """Load a list of structured data from the given file.
        File must be a .csv file. The first row of the file should be
        a header with the names of each column.

        Args:
            filename (os.path): The filename to load.

        Returns:
            List[Dict]: A list of rows, where each row contains {k : v}
               pairs for each field.

        Raises:
            ValueError: If the file is not a .csv file.
        """
        if not filename.endswith(".csv"):
            raise ValueError("File must be a CSV file.")

        documents = []
        with open(filename, "r", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            for row in reader:
                documents.append(row)
        return documents

    def _load_annotations(self, filename: os.path, anns_format: str):
        """Load a list of annotations from the given file.
        File must be a .json file. Each annotation must be in the correct
        format, e.g:

            [{
                "tokens": ["one", "three", "two"],
                "mentions": [
                    { "start": 0, "end": 1, "label": "number" },
                    { "start": 1, "end": 2, "label": "number" },
                    { "start": 2, "end": 3, "label": "number" }
                ],
                "relations": [
                    { "start": 1, "end": 0, "type": "bigger_than" },
                    { "start": 1, "end": 2, "type": "bigger_than" }
                ]
            }, ...]

        Args:
            filename (os.path): The filename to load.
            format (str): The format of the annotations. Can be either
               'quickgraph' or 'spert'.

        Returns:
            list: A list of Annotations.

        Raises:
            ValueError: If the file is not a JSON file, or fails to parse.
        """
        if not filename.endswith(".json"):
            raise ValueError("File must be a JSON file.")
        annotations = []
        with open(filename, "r", encoding="utf-8") as f:
            try:
                d = json.load(f)

                # If using quickgraph format, combine all the annotations
                # into one single list.
                if anns_format == "quickgraph":
                    anns = []
                    if isinstance(d, list):
                        anns = d
                    else:
                        if len(d.keys()) > 1:
                            logger.warning(
                                "Warning: you appear to be loading "
                                "multiple annotators' annotations."
                                "This may result in duplicate nodes - "
                                "see the readme for more details."
                            )
                        anns = []
                        for _, v in d.items():
                            anns += v
                else:
                    anns = d

                for ann in anns:
                    # Ignore unsaved annotations in QuickGraph
                    if (
                        anns_format == "quickgraph"
                        and "saved" in ann
                        and not ann["saved"]
                    ):
                        continue
                    normalise_annotation_format(ann, anns_format)

                    annotations.append(Annotation.from_dict(ann))

            except (JSONDecodeError, ValueError) as e:
                raise ValueError(
                    f"The .json file ({os.path.basename(filename)}) "
                    f"failed to parse due to the following issue: {e}"
                )
        logger.debug(f"Loaded {len(annotations)} annotations from {filename}.")
        return annotations

    def __repr__(self):
        """String representation of the dataset.

        Returns:
            str: String representation of the dataset.
        """
        return f"Dataset containing {len(self.documents)} documents."

    def to_list(self):
        """Return a list representation of this dataset.

        Returns:
            list[Dict]: A list of Dicts, where each Dict is one document from
               this dataset.
        """
        return [doc.to_dict() for doc in self.documents]


def _to_quickgraph(dataset: Dataset) -> List[Dict]:
    """Convert the given Dataset to a list of dicts compatible with Quickgraph.

    Args:
        dataset (Dataset): The dataset to convert.

    Returns:
        List[Dict]: A list of Quickgraph-compatible data.
    """
    quickgraph_docs = []
    for doc in dataset.documents:
        ann = doc.annotation

        entities = []
        relations = []

        for i, m in enumerate(ann.mentions):
            entities.append(
                {
                    "id": str(i + 1),
                    "start": m.start,
                    "end": m.end - 1,
                    "label": m.label,
                }
            )

        for i, r in enumerate(ann.relations):
            relations.append(
                {
                    "source_id": str(r.start.mention_id + 1),
                    "target_id": str(r.end.mention_id + 1),
                    "label": r.label,
                }
            )

        qd = {
            "original": " ".join(ann.tokens),
            "tokens": ann.tokens,
            "entities": entities,
            "relations": relations,
        }
        quickgraph_docs.append(qd)
    return quickgraph_docs


def _to_spert(dataset: Dataset) -> List[Dict]:
    """Convert the given Dataset to a list of dicts compatible with SPERT.

    Args:
        dataset (Dataset): The dataset to convert.

    Returns:
        List[Dict]: A list of SPERT-compatible data.
    """
    spert_docs = []
    for doc in dataset.documents:
        ann = doc.annotation

        entities = []
        relations = []

        for m in ann.mentions:
            entities.append(
                {
                    "start": m.start,
                    "end": m.end,
                    "type": m.label,
                }
            )

        for i, r in enumerate(ann.relations):
            relations.append(
                {
                    "head": r.start.mention_id,
                    "tail": r.end.mention_id,
                    "type": r.label,
                }
            )

        sd = {
            "tokens": ann.tokens,
            "entities": entities,
            "relations": relations,
        }
        spert_docs.append(sd)

    return spert_docs



"""A class that stores the annotations of a Document."""

from typing import List, Dict

from .Mention import Mention
from .Relation import Relation

from .size_limits import MAX_SENT_LENGTH, MAX_WORD_LENGTH

import logging as logger


class Annotation:
    """An Annotation for the textual portion of a Document.
    Contains a list of tokens, a list of mentions,
    a list of relations.

    :var tokens: A list of tokens (strings).
    :var mentions: A list of :class:`puggle.Mention.Mention` objects.
    :var relations: A list of :class:`puggle.Relation.Relation` objects.

    """

    def __init__(
        self,
        tokens: list[str],
        mentions: list[Dict],
        relations: list[Dict] = None,
    ):
        """Create a new Annotation.

        Args:
            tokens (list): The list of tokens of the document.
            mentions (list): The list of mentions of the document. Each
               mention must follow the correct format ('start', 'end',
               'label').
            relations (list, optional): The list of relations of the
               document. Each relation must follow the correct format
               ('start', 'end', 'type').
        """
        super().__init__()
        self.tokens = tokens
        self._mention_ids_map = {}
        self.mentions = self._parse_mentions(mentions, tokens)
        if relations:
            self.relations = self._parse_relations(relations)
        else:
            self.relations = []

    def to_dict(self):
        """Return a dictionary representation of this Annotation.
        Format will be similar to the input dataset.

        Returns:
            dict: Dictionary representation of this Annotation.
        """
        return {
            "tokens": self.tokens,
            "mentions": [m.to_dict() for m in self.mentions],
            "relations": [r.to_dict() for r in self.relations],
        }

    def _parse_mentions(self, mentions: List[Dict], tokens: List[Dict]):
        """Parse a list of mentions (which should each be stored as a dict).
        Ignore duplicate mentions i.e. mentions with the same start,
        end, and label.
        Each dict must have the format
           {"start": <int>, "end": <int>, "labels": List[str],
            "tokens": List[str]}

        Args:
            mentions (list): A list of dicts where each dict
              corresponds to a mention. Each dict must have 'start', 'end'
              and 'labels' keys.
            tokens (list): A list of tokens (i.e. the sentence).

        Returns:
            list: A list of Mention objects.
        """
        seen_mentions = set()
        mentions_list = []

        # Maintain a mapping from the mention id (in the original data)
        # to the id of the mention in the list of parsed Mentions,
        # so that relations can be correctly created.
        self._mention_ids_map = {}

        for i, m in enumerate(mentions):
            try:
                m_obj = Mention(
                    m["start"],
                    m["end"],
                    tokens[m["start"] : m["end"]],
                    m["label"],
                    i,
                )
                self._mention_ids_map[i] = m_obj
            except KeyError:
                raise KeyError(
                    "Could not parse document due to "
                    "missing keys. The 'mentions' must have 'start', 'end', "
                    "and 'label'."
                )

            if str(m_obj) not in seen_mentions:
                mentions_list.append(m_obj)
            seen_mentions.add(str(m_obj))

        return mentions_list

    def _parse_relations(self, relations: List[Dict]) -> List[Relation]:
        """Parse a given list of relations (which each should be a dict)
        and return a list of Relation objects.

        Args:
            relations (list): A list of relations. Each relation should
               have the keys 'start', 'end', and 'type'.

        Returns:
            List[Relation]: The parsed list of Relations.
        """

        relations_list = []
        _mention_ids_map = self._mention_ids_map
        for r in relations:
            start_key = "start"
            end_key = "end"
            try:
                r_obj = Relation(
                    _mention_ids_map[r[start_key]],
                    _mention_ids_map[r[end_key]],
                    r["type"],
                )
                relations_list.append(r_obj)
            except KeyError as e:
                raise KeyError(
                    f"Could not parse relations of document "
                    "because the mention corresponding to the relation with "
                    f"start: {r['start']} and end: {r['end']} was not found."
                )
            except ValueError as e:
                raise ValueError(
                    f"Could not parse relations of document "
                    f"due to the following error:\n{e}"
                )

        return relations_list

    @staticmethod
    def from_dict(d: dict):
        """Create an Annotation from a dictionary.

        Args:
            d (dict): The dictionary. Must contain tokens, mentions, relations.

        Returns:
            Annotation: An Annotation.

        Raises:
            ValueError: If the dictionary is missing a required
            key.
        """

        # If any required keys are missing, raise an error.
        if "tokens" not in d or "entities" not in d or "relations" not in d:
            raise ValueError(
                "Dictionary must contain tokens, entities, and relations."
            )

        # Validate things such as max word length, sent length, etc,
        # before attempting to create Annotation object
        if any(len(t) > MAX_WORD_LENGTH for t in d["tokens"]):
            word = ""
            for t in d["tokens"]:
                if len(t) > MAX_WORD_LENGTH:
                    word = t
            raise ValueError(
                f"Word must be at most {MAX_WORD_LENGTH} characters "
                f"long: {word}"
            )
        if len(d["tokens"]) > MAX_SENT_LENGTH:
            raise ValueError(
                f"Sentence must contain at most {MAX_SENT_LENGTH} words."
            )

        annotation = Annotation(
            tokens=d["tokens"],
            mentions=d["entities"],
            relations=d["relations"],
        )

        return annotation

    def __repr__(self):
        """Represent this document as a string.

        Returns:
            str: String representation of this document.
        """
        mens_str = (
            "\n  ".join([str(m) for m in self.mentions])
            if self.mentions
            else "-"
        )
        rels_str = (
            "\n  ".join([str(r) for r in self.relations])
            if self.relations
            else "-"
        )
        return "Tokens: %s\nMentions:  \n  %s\nRelations: \n  %s\n" % (
            " ".join(self.tokens),
            mens_str,
            rels_str,
        )



"""A class representing a single Document.
Contains an optional Annotation, and a list of fields."""

from typing import List, Dict

from .Annotation import Annotation


class Document:

    """A class representing a single Document.
    Contains an optional Annotation, and a list of fields.
    """

    def __init__(
        self,
        structured_fields: List[Dict] = None,
        annotation: Annotation = None,
    ):
        """Create a new document.

        Args:
            structured_fields (List[Dict], optional): List of fields.
            annotation (Annotation, optional): The Annotation of the textual
               part of this document (such as annotations over the short text)
        """
        super().__init__()
        self.fields = structured_fields
        self.annotation = annotation

    def to_dict(self):
        """Return a dict of this Document.

        Returns:
            dict: A dictionary representing this document.
        """
        return {
            "fields": self.fields,
            "annotations": self.annotation.to_dict()
            if self.annotation
            else None,
        }

    def __str__(self):
        return str(self.to_dict())



"""A nice logger."""
import sys
from loguru import logger

logger.remove()
logger.add(
    sys.stderr,
    colorize=True,
    format="<level>{level}</level>\t{message}",
    filter="puggle",
    level="INFO",
)
